/*

NAME
  mycp - copy with user abort

SYNOPSYS
  mycp fromfile tofile
  (for internal use by tbackup)

DESCRIPTION
  Copies fromfile to tofile.
  Aborts with exit code 1 of a write error occurs.
  Aborts with exit code 1 if the user presses `i' or `I'.

  Occasionaly syncs the floppy device, this makes for more efficient
  cache use and allows early user interruption in case of write
  errors.  With older linux library versions the sync syncs /all/
  devices (see fsync manual page) which is terribly inefficient.  But
  I doubt many people are still using these older libs now.

AUTHOR
  Koen Holtman, koen@blade.stack.urc.tue.nl   (kh1994)
*/

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <termios.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <linux/fd.h>
#include <sys/ioctl.h>

 void usage();
 void doit(); 
 
 void handler(int i); 
 void setstuff();
 void resetstuff();

/* fsync the floppy every 150 Kb. */
#define SYNCBLOCK (150*1024)

 char *pnam;
 char *infile;
 char *outfile;

 int child=0;

 int f=0,t=0; /* from and to files */


   
int main(int argc, char *argv[])
{ 
 pnam=argv[0];
 if(argc!=3) usage();

 infile=argv[1];
 outfile=argv[2];

 setstuff();

 doit();

 resetstuff();
 return 0;
}

void handler(int i) 
{ 
  fprintf(stderr,"Signal caught. Aborting...\n");
  resetstuff();
  exit(1);
}

struct termios tty;
int fflags;

void setstuff()
{
 signal(SIGINT,handler);

 /* Ugh. It took me far too long to find this. */
 tcgetattr(0, &tty);
 tty.c_lflag&=~ICANON;
 tcsetattr(0,TCSANOW, &tty);
 
 fflags=fcntl(0,F_GETFL);
 fflags|=O_NONBLOCK;
 fcntl(0,F_SETFL,(long)fflags);


} 

void resetstuff()
{
 tty.c_lflag|=ICANON;
 tcsetattr(0,TCSANOW, &tty);

 fflags&=~O_NONBLOCK;
 fcntl(0,F_SETFL,(long)fflags);

 if(child) waitpid(child, NULL, 0);
}


int apressed()
{
 char c; int r;

 while(1)
   {
     r=read(0,&c,1);
     if(r!=1) return 0;
     if((c=='i') || (c=='I')) return 1;
   }
}

void fdflush()
{
  if(t==0) return;
  if (ioctl(t,FDFLUSH,NULL) < 0) perror("mycp");
}

void checkabort()
{
 if(apressed())
   {
     fprintf(stderr,"\nInterrupting write.....\n");
     fflush(stderr);
     /* throw away unsynced writes. This helps a bit to reduce 
	time to exit */
     fdflush();
     
     resetstuff(); /* this also waits for the children */
     fdflush(); /* just to make sure */
     exit(1);
   }
}

void doit()
{
 int count,wcount,w; long total;
 char buf[4096];

 f=open(infile,O_RDONLY);
 if(f<0) 
   {
     fprintf(stderr,"%s: error opening %s\n",pnam,infile);
     resetstuff();
     exit(1);
   }

 t=creat(outfile,0644);
 if(t<0) 
   {
     fprintf(stderr,"%s: error opening %s\n",pnam,outfile);
     resetstuff();
     exit(1);
   }

 /* perform the copy. */
 total=0;

 while(1)
   {
     count=read(f,buf,4096);
     if(count<=0) break;

     checkabort();

     wcount=0;
     while(wcount<count)
       {
	 w=write(t,buf+wcount,count-wcount);
         if(w<=0)
            { fprintf(stderr,"%s: write error on %s!\n",pnam,outfile);
              resetstuff();
              exit(1);
            }
         wcount+=w;
       }

     total+=count;

     if(total>SYNCBLOCK) 
     {
	 /* we start a child to do the syncing, as fsync is a blocking
	    operation, and we want to read and write on.
	    The sync eventually ends up syncing a lot of the new writes too.
	    
	    We wait for the previous syncing child, else the writing would 
	    get out of sync with the syncing. 
	 */
	 if(child!=0)
	 {
            waitpid(child, NULL, 0);
	 }

	 checkabort();

         /* create a child to flush the floppy buffer */
         if((child=fork())==0)
	 {
	     fsync(t);
	     exit(0);
	 }

	 total=0;
     }
   }

 close(f);
 close(t);
}

void usage()
{
 fprintf(stderr,
  "usage: %s <file> <floppy>\n"
        ,pnam);
 exit(1);
}

